import numpy as np
import pytest

from eerily.generators.utils.base import ConstantIterator


@pytest.mark.parametrize(
    "constant, expected_val",
    [
        pytest.param(10, 10, id="constant_10"),
        pytest.param("a", "a", id="constant_a"),
        pytest.param([1, 2], [1, 2], id="constant_list_1_2"),
        pytest.param(np.array([1, 2]), np.array([1, 2]), id="constant_array_1_2"),
    ],
)
def test_constant_iterator(constant, expected_val):

    length = 10

    constant_iterator = ConstantIterator(constant=constant)

    results = [next(constant_iterator) for _ in range(length)]

    assert len(results) == length

    if isinstance(expected_val, np.ndarray):
        assert all([np.array_equal(i, expected_val) for i in results])
    else:
        assert all([i == expected_val for i in results])



import copy
from dataclasses import dataclass
from typing import Any, List

from eerily.generators.utils.stepper import BaseStepper, StepperParams


@dataclass(frozen=True)
class ConstStepperParams(StepperParams):
    """
    Parameters for constant model.

    ```python
    cp = ConstStepperParams(
        initial_state=["A"],
        variable_names=["name"],
    )
    ```
    """

    initial_state: List[Any]
    variable_names: List[Any]


class ConstantStepper(BaseStepper):
    """Generates constant values for an given initial condition.

    ```python
    csp = ConstStepperParams(initial_state=[1], variable_names=["y"])
    cs = ConstantStepper(model_params=csp)

    next(cs)
    ```
    """

    def compute_step(self):
        return dict(zip(self.model_params.variable_names, self.model_params.initial_state))

    def __repr__(self) -> str:
        return "ConstantStepper: \n" f"parameters: {self.model_params}\n" f"current_state: {self.current_state}"


@dataclass(frozen=True)
class SequenceStepperParams(StepperParams):
    """
    Parameters for constant model.

    ```python
    cp = SequenceStepperParams(
        initial_state=[1],
        variable_names=["step"],
        step_sizes=[1]
    )
    ```
    """

    initial_state: List[Any]
    variable_names: List[Any]
    step_sizes: List[Any]


class SequenceStepper(BaseStepper):
    """Generates sequence values for an given initial condition.

    ```python
    csp = SequenceStepperParams(
        initial_state=[1],
        variable_names=["y"],
        step_sizes=[1]
    )
    cs = SequenceStepper(model_params=csp)

    next(cs)
    ```
    """

    def compute_step(self):
        new_state = []
        for c, i in zip(self.current_state, self.model_params.step_sizes):
            new_state.append(c + i)

        self.current_state = copy.deepcopy(new_state)
        return dict(zip(self.model_params.variable_names, self.current_state))

    def __repr__(self) -> str:
        return "SequenceStepper: \n" f"parameters: {self.model_params}\n" f"current_state: {self.current_state}"



from typing import Optional

import numpy as np


class PoissonEvent:
    """A Poisson process event generator.

    See [this notebook](https://github.com/btel/python-in-neuroscience-tutorials/blob/master/poisson_process.ipynb) for more about a Poisson process.

    ```python
    rate = 0.1
    pe = PoissonEvents(rate=rate)
    next(pe)
    ```

    :param lambda: the mean rate of the Poisson process
    :param seed: seed of the RNG for reproducibility
    """

    def __init__(self, rate: float, seed: Optional[float] = None):
        self.rate = rate
        self.rng = np.random.default_rng(seed=seed)

    def __iter__(self):
        return self

    def __next__(self) -> float:
        random_state = self.rng.random() <= self.rate
        return int(random_state)



from typing import Optional

import numpy as np


class GaussianNoise:
    """1 D Gaussian noise

    :param mu: mean of the Gaussian distribution
    :param std: standard deviation of the Gaussian distribution
    :param seed: seed of the RNG for reproducibility
    """

    def __init__(self, mu: float, std: float, seed: Optional[float] = None):
        self.mu = mu
        self.std = std
        self.rng = np.random.default_rng(seed=seed)

    def __iter__(self):
        return self

    def __next__(self) -> float:
        return self.rng.normal(self.mu, self.std)


class LogNormalNoise:
    """1 D lognormal noise

    :param mu: mean of the Gaussian distribution
    :param std: standard deviation of the Gaussian distribution
    :param seed: seed of the RNG for reproducibility
    """

    def __init__(self, mu: float, std: float, seed: Optional[float] = None):
        self.mu = mu
        self.std = std
        self.rng = np.random.default_rng(seed=seed)

    def __iter__(self):
        return self

    def __next__(self) -> float:
        return self.rng.lognormal(self.mu, self.std)


class MultiGaussianNoise:
    """A multivariate Gaussian noise

    To generate constants,

    ```python
    mge = MultiGaussianEpsilon(
        mu=np.array([1,2]), cov=np.array([
            [0, 0],
            [0, 0]
        ])
    )
    ```

    To generate independent noises,

    ```python
    mge = MultiGaussianEpsilon(
        mu=np.array([1,2]), cov=np.array([
            [1, 0],
            [0, 1]
        ])
    )
    ```

    :param mu: means of the variables
    :param cov: covariance of the variables
    :param seed: seed of the random number generator for reproducibility
    """

    def __init__(self, mu: np.ndarray, cov: np.ndarray, seed: Optional[float] = None):
        self.mu = mu
        self.cov = cov
        self.rng = np.random.default_rng(seed=seed)

    def __next__(self) -> np.ndarray:
        return self.rng.multivariate_normal(self.mu, self.cov)



from typing import Any


class ConstantIterator:
    """An iterator that emits constant values.

    ```python
    pe = ConstantIterator(constant=1)
    next(pe)
    ```

    :param constant: the constant value to be emmited.
    """

    def __init__(self, constant: Any):
        self.constant = constant

    def __iter__(self):
        return self

    def __next__(self) -> Any:
        return self.constant



from typing import Any, Optional, Sequence, Tuple, Union

import numpy as np


class Choices:
    """Generating data by choosing from some given values.

    ```python
    elements = [0,1]
    c = Choices(elements=elements)
    next(c)
    ```

    :param elements: the elements to choose from
    :param seed: seed of the RNG for reproducibility
    """

    def __init__(
        self,
        elements: Sequence[Any],
        seed: Optional[float] = None,
    ):
        self.elements = elements
        self.indices = range(len(elements))

        self.rng = np.random.default_rng(seed=seed)

    def __next__(self) -> Any:
        idx = self.rng.choice(self.indices)
        return self.elements[idx]



import copy
from dataclasses import dataclass
from typing import Dict, Iterator

from eerily.generators.utils.stepper import BaseStepper, StepperParams


@dataclass(frozen=True)
class SpikingEventParams(StepperParams):
    """
    Parameters for spiking events.

    :param spike: the spiking process, e.g., Poisson process using
        [`PoissonEvents`][eerily.generators.utils.events.PoissonEvents].
    :param spike_level: the level of spikes, e.g.
        [`GaussianNoise`][eerily.generators.utils.noise.GaussianNoise]
        with some positve mean. This parameter determines the height of the spikes.
    :param background: the stochastic noise level, e.g.
        [`GaussianNoise`][eerily.generators.utils.noise.GaussianNoise].
    """

    spike: Iterator
    spike_level: Iterator
    background: Iterator

    def __post_init__(self):
        if self.initial_state is None:
            self.initial_state = 0
        if self.variable_names is None:
            self.variable_names = ["event"]


class SpikingEventStepper(BaseStepper):
    """Calculates the next step in a spiking event.

    :param model_params: a dataclass that contains the necessary parameters for the model.
        e.g., [`SpikingEventParams`][eerily.generators.spiking.SpikingEventParams]
    """

    def compute_step(self) -> Dict[str, float]:

        background = next(self.model_params.background)  # type: ignore
        spike = next(self.model_params.spike)  # type: ignore
        spike_level = next(self.model_params.spike_level)  # type: ignore

        v_next = background + spike * spike_level

        self.current_state = v_next

        return copy.deepcopy(self.current_state)



import copy
from dataclasses import dataclass
from typing import Dict, Iterator, Optional, Sequence, Union

import pandas as pd
from loguru import logger

from eerily.generators.utils.stepper import BaseStepper, StepperParams


@dataclass(frozen=True)
class LinearElasticityParams(StepperParams):
    """
    Parameters for constant elasticity model.

    ```python
    length = 10

    lep = LinearElasticityParams(
        initial_state={"log_demand": 3, "log_price": 0.5, "elasticity": None},
        log_prices=iter(range(length)),
        elasticity=iter([-3] * length),
        variable_names=["log_demand", "log_price", "elasticity"],
    )
    ```

    !!! warning "Initial Condition"
        Initial condition is a dictionary with at least two keys `sale` and `price`.

        Note that the initial condition is NOT returned in the iterator.


    :param elasticity: an iterator that generates the elasticity to be used for each step
    :param log_prices: an iterator that generates the log prices in each step
    """

    elasticity: Iterator
    log_prices: Iterator
    log_base_demand: Optional[Iterator] = None

    def __post_init__(self):
        if self.initial_state is None:
            self.initial_state = {"log_demand": 1, "log_price": 10, "elasticity": None}
        if self.variable_names is None:
            self.variable_names = ["log_demand", "log_price", "elasticity"]


class ElasticityStepper(BaseStepper):
    """Generates the next time step for an given initial condition.

    We use the following formula to generate the data

    $$
    \ln Q' = \ln Q + \epsilon (\ln P' - \ln P)
    $$

    Define new log transformed variables to make this a linear relation

    $$
    y' = y + \epsilon (x' - x).
    $$

    For example, with initial condition

    ```
    initial_condition = {"log_price": 1, "log_sales": 10, "elasticity": None}
    ```

    For a deterministic model, we have

    ```python
    length = 10
    elasticity = iter([-3] * length)
    log_prices = iter(range(length))

    initial_condition = {"log_demand": 3, "log_price": 0.5, "elasticity": None}

    lep = LinearElasticityParams(
        initial_state=initial_condition,
        log_prices=log_prices,
        elasticity=elasticity,
        variable_names=["log_demand", "log_price", "elasticity"],
    )

    es = ElasticityStepper(model_params=lep)

    next(es)
    ```

    We have utils in [`eerily.generators.utils`][eerily.generators.utils]
    to help the user creating elasticty and log prices generators.
    For example, we can create a constant iterator using
    [`ConstantIterator`][eerily.generators.utils.base.ConstantIterator]


    """

    def compute_step(self):

        elasticity = next(self.model_params.elasticity)

        current_log_price = self.current_state["log_price"]
        current_log_demand = self.current_state["log_demand"]

        next_log_price = next(self.model_params.log_prices)

        if self.model_params.log_base_demand is None:
            next_log_demand = current_log_demand + elasticity * (next_log_price - current_log_price)
        else:
            next_log_base_demand = next(self.model_params.log_base_demand)
            next_log_demand = next_log_base_demand + elasticity * next_log_price
            self.current_state["log_base_demand"] = next_log_base_demand

        self.current_state["log_demand"] = next_log_demand
        self.current_state["log_price"] = next_log_price
        self.current_state["elasticity"] = elasticity

        return copy.deepcopy(self.current_state)

    def __repr__(self) -> str:
        return "ElasticityStepper: \n" f"parameters: {self.model_params}\n" f"current_state: {self.current_state}"



import copy
from dataclasses import dataclass
from typing import Dict, Iterator

from eerily.generators.utils.stepper import BaseStepper, StepperParams


@dataclass(frozen=True)
class BrownianMotionParams(StepperParams):
    """
    Parameters for Brownian motion

    :param gamma: the damping factor $\gamma$ of the Brownian motion.
    :param delta_t: the minimum time step $\Delta t$.
    :param force_densities: the stochastic force densities, e.g.
        [`GaussianNoise`][eerily.generators.utils.noise.GaussianNoise].
    :param initial_state: the initial velocity $v(0)$.
    :param variable_names: variable names of the given initial condition
    """

    gamma: float
    delta_t: float
    force_densities: Iterator


class BrownianMotionStepper(BaseStepper):
    """Calculates the next step in a brownian motion.

    ??? note "Brownian Motion"

        Macroscopically, Brownian Motion can be described by the notion of random forces on the particles,

        $$\\frac{d}{dt} v(t) + \gamma v(t) = R(t),$$

        where $v(t)$ is the velocity at time $t$ and $R(t)$ is the stochastic force density from the reservoir particles.

        To simulate it numerically, we rewrite

        $$\\frac{d}{dt} v(t) + \gamma v(t) = R(t),$$

        as

        $$\Delta v (t+1) = R(t) \Delta t - \gamma v(t) \Delta t$$


    !!! example "Example Code"

        ```python
        guassian_force = GaussianForce(mu=0, std=1, seed=seed)
        bm_params = BrownianMotionParams(
            gamma=0, delta_t=0.1, force_densities=guassian_force, initial_state=np.array([0]),
            variable_names=["v"]
        )

        bms = BrownianMotionStepper(
            model_params = bm_params
        )

        next(bms)
        ```

    :param model_params: a dataclass that contains the necessary parameters for the model.
        e.g., [`BrownianMotionParams`][eerily.generators.brownian.BrownianMotionParams]
    """

    def compute_step(self) -> Dict[str, float]:

        force_density = next(self.model_params.force_densities)  # type: ignore

        v_next = (
            self.current_state
            + force_density * self.model_params.delta_t  # type: ignore
            - self.model_params.gamma * self.current_state * self.model_params.delta_t  # type: ignore
        )

        self.current_state = v_next

        return copy.deepcopy(self.current_state)



import math
from functools import cached_property
from typing import Dict, List


class Pendulum:
    """Class for generating time series data for a pendulum.

    :param length: Length of the pendulum.
    :param gravity: Acceleration due to gravity.
    """

    def __init__(self, length: float, gravity: float = 9.81) -> None:
        self.length = length
        self.gravity = gravity

    @cached_property
    def period(self) -> float:
        """Calculate the period of the pendulum."""
        return 2 * math.pi * math.sqrt(self.length / self.gravity)

    def __call__(
        self, num_periods: int, num_samples_per_period: int, initial_angle: float = 0.1
    ) -> Dict[str, List[float]]:
        """Generate time series data for the pendulum.

        Returns a list of floats representing the angle
        of the pendulum at each time step.

        :param num_periods: Number of periods to generate.
        :param num_samples_per_period: Number of samples per period.
        :param initial_angle: Initial angle of the pendulum.
        """
        time_step = self.period / num_samples_per_period
        steps = []
        time_series = []
        for i in range(num_periods * num_samples_per_period):
            t = i * time_step
            # angle = self.length * math.cos(2 * math.pi * t / self.period)
            angle = initial_angle * math.cos(2 * math.pi * t / self.period)
            steps.append(t)
            time_series.append(angle)

        return {"t": steps, "theta": time_series}



import copy
from dataclasses import dataclass
from typing import Iterator

import numpy as np

from eerily.generators.utils.stepper import BaseStepper, StepperParams


@dataclass(frozen=True)
class ARModelParams(StepperParams):
    """Parameters of our AR model,

    $$s(t+1) = \phi_0 + \phi_1 s(t) + \epsilon.$$

    :param delta_t: step size of time in each iteration
    :param phi0: pho_0 in the AR model
    :param phi1: pho_1 in the AR model
    :param epsilon: noise iterator, e.g., Gaussian noise
    :param initial_state: a dictionary of the initial state,
        e.g., `np.array([1])`
    :param variable_name: variable names
    """

    delta_t: float
    phi0: float
    phi1: float
    epsilon: Iterator


class AR1Stepper(BaseStepper):
    """Stepper that calculates the next step in time in an AR model

    :param model_params: parameters for the AR model
    """

    def compute_step(self):
        epsilon = next(self.model_params.epsilon)

        next_s = self.model_params.phi0 + self.model_params.phi1 * self.current_state + epsilon
        self.current_state = next_s

        return copy.deepcopy(self.current_state)


@dataclass(frozen=True)
class VAR1ModelParams(StepperParams):
    r"""Parameters of our VAR model,

    $$
    \begin{equation}
    \begin{pmatrix}s^{(1)}(t+1) \\ s^{(2)}(t+1) \end{pmatrix} =
    \begin{pmatrix} \phi^{(1)}_0 \\ \phi^{(2)}_0 \end{pmatrix} +
    \begin{pmatrix}\phi_{1, 11} & \phi_{1, 12}\\ \phi_{1, 21} & \phi_{1, 22} \end{pmatrix}
    \begin{pmatrix}s^{(1)}(t) \\ s^{(2)}(t) \end{pmatrix} +
    \begin{pmatrix}\epsilon^{(1)} \\ \epsilon^{(2)} \end{pmatrix}.
    \end{equation}
    $$

    :param delta_t: step size of time in each iteration
    :param phi0: pho_0 in the AR model
    :param phi1: pho_1 in the AR model
    :param epsilon: noise iterator, e.g., Gaussian noise
    :param initial_state: an array of the initial state, e.g., `{"s": 1}`
    """

    delta_t: float
    phi0: np.ndarray
    phi1: np.ndarray
    epsilon: Iterator


class VAR1Stepper(BaseStepper):
    """Calculate the next values using VAR(1) model.

    :param model_params: the parameters of the VAR(1) model, e.g.,
        [`VAR1ModelParams`][eerily.generators.var.VAR1ModelParams]
    """

    def compute_step(self):

        epsilon = next(self.model_params.epsilon)
        phi0 = self.model_params.phi0
        phi1 = self.model_params.phi1

        self.current_state = phi0 + np.matmul(phi1, self.current_state) + epsilon

        return copy.deepcopy(self.current_state)



from typing import Iterator, Literal, Optional


class Factory:
    """
    An experimental factor that generates a history
    based on a stepper and the given length.
    """

    def __init__(self, format: Optional[Literal["list", "dataframe"]] = None):
        self.format = format

    def __call__(self, stepper: Iterator, length: int):
        i = 0
        while i < length:
            yield next(stepper)
            i += 1



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, List, Optional, Union

from loguru import logger


class StepperOperator:
    """Allowing `&` and `+` for the steppers."""

    def __add__(
        self, another_stepper: Union[StepperOperator, BaseStepper, SequentialStepper, MergedStepper]
    ) -> SequentialStepper:
        return SequentialStepper([self, another_stepper])

    def __and__(
        self, another_stepper: Union[StepperOperator, BaseStepper, SequentialStepper, MergedStepper]
    ) -> MergedStepper:
        return MergedStepper([self, another_stepper])


@dataclass(frozen=True)
class StepperParams:
    """Base Parameters for Stepper

    :param initial_state: the initial state, e.g., `np.array([1])`
    :param variable_name: variable names of the time series provided as a list.
    """

    initial_state: Any
    variable_names: List[Any]


class BaseStepper(ABC, StepperOperator):
    """A framework to evolve a DGP to the next step"""

    def __init__(self, model_params: StepperParams, length: Optional[int] = None) -> None:
        self.model_params = model_params
        self.current_state = copy.deepcopy(self.model_params.initial_state)
        self.length = length
        self._counter = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.length is None:
            logger.warning("length is not set")
            self._counter += 1
            return self.compute_step()
        else:
            if self._counter < self.length:
                self._counter += 1
                return self.compute_step()
            else:
                raise StopIteration

    def __len__(self):
        return self.length

    def get_iterator(self):
        return self.__iter__()

    def __str__(self) -> str:
        return f"Model Parameters: {self.model_params}"

    @abstractmethod
    def compute_step(self):
        pass


class SequentialStepper(StepperOperator):
    def __init__(
        self,
        iterators: List[Union[StepperOperator, BaseStepper, SequentialStepper, MergedStepper]],
    ):
        self.iterators: List[BaseStepper] = []
        self._length = 0
        for stepper in iterators:
            if isinstance(stepper, SequentialStepper):
                self.iterators.extend(stepper.iterators)
            elif isinstance(stepper, BaseStepper):
                self.iterators.append(stepper)
            else:
                raise TypeError("Please provide a list of steppers")

    def __iter__(self):
        for stepper in self.iterators:
            for _ in range(stepper.length):
                yield next(stepper)

    @property
    def length(self):
        self._length = sum([stepper.length for stepper in self.iterators])
        return self._length

    def __len__(self):
        return self.length


class MergedStepper(StepperOperator):
    def __init__(
        self,
        iterators: List[Union[StepperOperator, BaseStepper, SequentialStepper, MergedStepper]],
    ):
        self.iterators: List[BaseStepper] = []
        self._length = 0
        self._counter = 0
        for stepper in iterators:
            if isinstance(stepper, MergedStepper):
                self.iterators.extend(stepper.iterators)
            elif isinstance(stepper, BaseStepper):
                self.iterators.append(stepper)
            else:
                raise TypeError("Please provide a list of steppers")

    def __iter__(self):
        if all([stepper.length is not None for stepper in self.iterators]):
            length = min([stepper.length for stepper in self.iterators])
        else:
            raise ValueError("length is not set")

        for vals in zip(range(length), *self.iterators):
            idx = vals[0]
            iter_values = vals[1:]
            combined = {}
            for val in iter_values:
                if isinstance(val, dict):
                    combined.update(val)
                else:
                    raise NotImplementedError("Please implement __and__ for your steppers")
            yield combined
            self._counter += idx

    @property
    def length(self):
        self._length = min([stepper.length for stepper in self.iterators])
        return self._length

    def __len__(self):
        return self.length



import numpy as np
import pandas as pd
import pytest

from eerily.generators.elasticity import ElasticityStepper, LinearElasticityParams
from eerily.generators.utils.base import ConstantIterator
from eerily.generators.utils.noises import GaussianNoise


@pytest.fixture
def seed():
    return 42


@pytest.fixture
def length():
    return 10


@pytest.fixture
def constant_elasticity():
    return ConstantIterator(constant=-3)


@pytest.fixture
def log_prices(length):
    return iter(range(length))


@pytest.fixture
def log_base_demand(length):
    return iter(range(length))


@pytest.fixture
def deterministic_elasticity_stepper(constant_elasticity, log_prices):

    initial_condition = {"log_demand": 3, "log_price": 0.5, "elasticity": None}

    lep = LinearElasticityParams(
        initial_state=initial_condition,
        log_prices=log_prices,
        elasticity=constant_elasticity,
        variable_names=["log_demand", "log_price", "elasticity"],
    )

    return ElasticityStepper(model_params=lep)


@pytest.fixture
def deterministic_base_demand_elasticity_stepper(constant_elasticity, log_prices, log_base_demand):

    initial_condition = {"log_demand": 3, "log_price": 0.5, "elasticity": None}

    lep = LinearElasticityParams(
        initial_state=initial_condition,
        log_prices=log_prices,
        elasticity=constant_elasticity,
        log_base_demand=log_base_demand,
        variable_names=["log_demand", "log_price", "elasticity"],
    )

    return ElasticityStepper(model_params=lep)


def test_deterministic_elasticity_stepper(deterministic_elasticity_stepper, length):

    container = []
    for _ in range(length):
        container.append(next(deterministic_elasticity_stepper))

    container_truth = [
        {"log_price": 0, "log_demand": 4.5, "elasticity": -3},
        {"log_price": 1, "log_demand": 1.5, "elasticity": -3},
        {"log_price": 2, "log_demand": -1.5, "elasticity": -3},
        {"log_price": 3, "log_demand": -4.5, "elasticity": -3},
        {"log_price": 4, "log_demand": -7.5, "elasticity": -3},
        {"log_price": 5, "log_demand": -10.5, "elasticity": -3},
        {"log_price": 6, "log_demand": -13.5, "elasticity": -3},
        {"log_price": 7, "log_demand": -16.5, "elasticity": -3},
        {"log_price": 8, "log_demand": -19.5, "elasticity": -3},
        {"log_price": 9, "log_demand": -22.5, "elasticity": -3},
    ]

    assert container == container_truth


def test_deterministic_base_demand_elasticity_stepper(deterministic_base_demand_elasticity_stepper, length):

    container = []
    for _ in range(length):
        container.append(next(deterministic_base_demand_elasticity_stepper))

    container_truth = [
        {"log_demand": 0, "log_price": 0, "elasticity": -3, "log_base_demand": 0},
        {"log_demand": -2, "log_price": 1, "elasticity": -3, "log_base_demand": 1},
        {"log_demand": -4, "log_price": 2, "elasticity": -3, "log_base_demand": 2},
        {"log_demand": -6, "log_price": 3, "elasticity": -3, "log_base_demand": 3},
        {"log_demand": -8, "log_price": 4, "elasticity": -3, "log_base_demand": 4},
        {"log_demand": -10, "log_price": 5, "elasticity": -3, "log_base_demand": 5},
        {"log_demand": -12, "log_price": 6, "elasticity": -3, "log_base_demand": 6},
        {"log_demand": -14, "log_price": 7, "elasticity": -3, "log_base_demand": 7},
        {"log_demand": -16, "log_price": 8, "elasticity": -3, "log_base_demand": 8},
        {"log_demand": -18, "log_price": 9, "elasticity": -3, "log_base_demand": 9},
    ]

    assert container == container_truth


@pytest.fixture
def gaussian_elasticity(seed):
    elasticity_mean = -3
    elasticity_std = 0.5
    return GaussianNoise(mu=elasticity_mean, std=elasticity_std, seed=seed)


@pytest.fixture
def stochastic_elasticity_stepper(gaussian_elasticity, log_prices):

    initial_condition = {"log_demand": 3, "log_price": 0.5, "elasticity": None}

    lep = LinearElasticityParams(
        initial_state=initial_condition,
        log_prices=log_prices,
        elasticity=gaussian_elasticity,
        variable_names=["log_demand", "log_price", "elasticity"],
    )

    return ElasticityStepper(model_params=lep)


def test_stochastic_elasticity_stepper(stochastic_elasticity_stepper, length):

    container = []
    for _ in range(length):
        container.append(next(stochastic_elasticity_stepper))

    container_truth = [
        {
            "log_price": 0,
            "log_demand": 4.423820730061392,
            "elasticity": -2.847641460122784,
        },
        {
            "log_price": 1,
            "log_demand": 0.9038286769411439,
            "elasticity": -3.519992053120248,
        },
        {
            "log_price": 2,
            "log_demand": -1.7209457251556275,
            "elasticity": -2.6247744020967714,
        },
        {
            "log_price": 3,
            "log_demand": -4.250663366960021,
            "elasticity": -2.529717641804393,
        },
        {
            "log_price": 4,
            "log_demand": -8.22618096128694,
            "elasticity": -3.9755175943269183,
        },
        {
            "log_price": 5,
            "log_demand": -11.877270714718097,
            "elasticity": -3.651089753431159,
        },
        {
            "log_price": 6,
            "log_demand": -14.813350513134456,
            "elasticity": -2.9360797984163574,
        },
        {
            "log_price": 7,
            "log_demand": -17.971471809306248,
            "elasticity": -3.158121296171791,
        },
        {
            "log_price": 8,
            "log_demand": -20.979872388058393,
            "elasticity": -3.0084005787521444,
        },
        {
            "log_price": 9,
            "log_demand": -24.406394351845183,
            "elasticity": -3.42652196378679,
        },
    ]

    pd.testing.assert_frame_equal(pd.DataFrame(container), pd.DataFrame(container_truth), check_like=True)



import pytest

from eerily.generators.utils.stepper import BaseStepper, StepperParams


def test_base_stepper():
    class DummyStepperParams(StepperParams):
        pass

    class DummyStepper(BaseStepper):
        def compute_step(self):
            return 1

    stepper_params = DummyStepperParams(initial_state=1, variable_names=["y"])
    stepper = DummyStepper(model_params=stepper_params)

    assert next(stepper) == 1


@pytest.mark.parametrize(
    "length",
    [
        pytest.param(1, id="length_1"),
        pytest.param(5, id="length_5"),
        pytest.param(10, id="length_10"),
    ],
)
def test_base_stepper_with_length(length):
    class DummyStepperParams(StepperParams):
        pass

    class DummyStepper(BaseStepper):
        def compute_step(self):
            return 1

    stepper_params = DummyStepperParams(initial_state=1, variable_names=["y"])
    stepper = DummyStepper(model_params=stepper_params, length=length)

    assert list(stepper) == [1] * length


def test_base_stepper_add():
    class DummyStepperParams(StepperParams):
        pass

    class DummyStepper(BaseStepper):
        def compute_step(self):
            return dict(zip(self.model_params.variable_names, self.model_params.initial_state))

    stepper_params_1 = DummyStepperParams(initial_state=[1], variable_names=["y1"])
    stepper_1 = DummyStepper(model_params=stepper_params_1, length=3)

    stepper_params_2 = DummyStepperParams(initial_state=[2], variable_names=["y2"])
    stepper_2 = DummyStepper(model_params=stepper_params_2, length=2)

    generator = stepper_1 + stepper_2

    assert list(generator) == [{"y1": 1}] * 3 + [{"y2": 2}] * 2


def test_base_stepper_add_3():
    class DummyStepperParams(StepperParams):
        pass

    class DummyStepper(BaseStepper):
        def compute_step(self):
            return dict(zip(self.model_params.variable_names, self.model_params.initial_state))

    stepper_params_1 = DummyStepperParams(initial_state=[1], variable_names=["y"])
    stepper_1 = DummyStepper(model_params=stepper_params_1, length=3)

    stepper_params_2 = DummyStepperParams(initial_state=[2], variable_names=["y"])
    stepper_2 = DummyStepper(model_params=stepper_params_2, length=2)

    stepper_params_3 = DummyStepperParams(initial_state=[3], variable_names=["y"])
    stepper_3 = DummyStepper(model_params=stepper_params_3, length=2)

    generator = stepper_1 + stepper_2 + stepper_3

    assert list(generator) == [{"y": 1}] * 3 + [{"y": 2}] * 2 + [{"y": 3}] * 2


def test_base_stepper_and():
    class DummyStepperParams(StepperParams):
        pass

    class DummyStepper(BaseStepper):
        def compute_step(self):
            return dict(zip(self.model_params.variable_names, self.model_params.initial_state))

    stepper_params_1 = DummyStepperParams(initial_state=[1], variable_names=["y1"])
    stepper_1 = DummyStepper(model_params=stepper_params_1, length=5)

    stepper_params_2 = DummyStepperParams(initial_state=[2], variable_names=["y2"])
    stepper_2 = DummyStepper(model_params=stepper_params_2, length=4)

    generator = stepper_1 & stepper_2

    assert list(generator) == [{"y1": 1, "y2": 2}] * 4


def test_base_stepper_and_3():
    class DummyStepperParams(StepperParams):
        pass

    class DummyStepper(BaseStepper):
        def compute_step(self):
            return dict(zip(self.model_params.variable_names, self.model_params.initial_state))

    stepper_params_1 = DummyStepperParams(initial_state=[1], variable_names=["y1"])
    stepper_1 = DummyStepper(model_params=stepper_params_1, length=5)

    stepper_params_2 = DummyStepperParams(initial_state=[2], variable_names=["y2"])
    stepper_2 = DummyStepper(model_params=stepper_params_2, length=4)

    stepper_params_3 = DummyStepperParams(initial_state=[3], variable_names=["y3"])
    stepper_3 = DummyStepper(model_params=stepper_params_3, length=4)

    generator = stepper_1 & stepper_2 & stepper_3

    assert list(generator) == [{"y1": 1, "y2": 2, "y3": 3}] * 4



import math

from ts_bolt.physics.pendulum import Pendulum


def test_pendulum_generate_data():  # type: ignore
    pendulum = Pendulum(length=1.0, gravity=9.81)

    assert pendulum.length == 1.0
    assert pendulum.gravity == 9.81
    assert pendulum.period == 2 * math.pi * math.sqrt(pendulum.length / pendulum.gravity)

    num_periods = 2
    num_samples_per_period = 10
    data = pendulum(num_periods, num_samples_per_period)

    assert len(data["t"]) == num_periods * num_samples_per_period



from eerily.generators.naive import (
    ConstantStepper,
    ConstStepperParams,
    SequenceStepper,
    SequenceStepperParams,
)


def test_constant_stepper():
    csp = ConstStepperParams(initial_state=[1], variable_names=["y"])
    cs = ConstantStepper(model_params=csp)

    assert next(cs) == {"y": 1}


def test_constant_stepper_length():
    csp = ConstStepperParams(initial_state=[1], variable_names=["y"])
    cs = ConstantStepper(model_params=csp, length=3)

    assert list(cs) == [{"y": 1}] * 3


def test_sequence_stepper():
    ssp = SequenceStepperParams(initial_state=[1], variable_names=["y"], step_sizes=[1])
    ss = SequenceStepper(model_params=ssp)

    assert next(ss) == {"y": 2}


def test_sequence_stepper_length():
    ssp = SequenceStepperParams(initial_state=[1], variable_names=["y"], step_sizes=[1])
    ss = SequenceStepper(model_params=ssp, length=5)

    assert list(ss) == [{"y": i} for i in range(2, 7)]
